#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-12-04
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

from worker.ftp_client import FTPClient
from pyftpdlib.authorizers import DummyAuthorizer
from pyftpdlib.handlers import FTPHandler
from pyftpdlib.servers import ThreadedFTPServer
from threading import Thread
from unittest import TestCase
import _thread
import os
import sys
import time
import shutil

cliiii = None
client = None
server = None


class TestFTPClient(TestCase, Thread):
    def setup_class(self):
        # init data
        if os.path.isdir("./tmp"):
            shutil.rmtree("./tmp")
        os.mkdir("./tmp")
        os.mkdir("./tmp/dir0")
        open("./tmp/file0", "w").close()
        f = open("./tmp/dir0/file0", "w")
        f.write("123456")
        f.close()
        # init client and server
        global cliiii
        global client
        global server
        # init client
        client = FTPClient("localhost", 21, 1, "root", "root", 1024)
        cliiii = FTPClient("localhost", 21, 1, "hyb", "hyb", 1024)
        # init server
        auth = DummyAuthorizer()
        auth.add_user("root", "root", "./tmp", perm="elradfmwMT")
        auth.add_user("hyb", "hyb", "./tmp", perm="elr")
        handler = FTPHandler
        handler.authorizer = auth
        server = ThreadedFTPServer(("localhost", 21), handler)
        _thread.start_new_thread(server.serve_forever, ())
        time.sleep(0.001)

    def teardown_class(self):
        server.close_all()

    def test_file_info(self):
        # root user
        assert client.isexist("")
        assert client.isdir("")
        assert not client.isfile("")
        assert not client.isexist("1/2/3")
        assert not client.isdir("1/2/3")
        assert not client.isfile("1/2/3")
        assert client.isexist("dir0")
        assert client.isexist("file0")
        assert client.isexist("dir0/file0")
        assert client.isdir("dir0")
        assert client.isfile("file0")
        assert client.isfile("dir0/file0")
        # hyb user
        assert cliiii.isexist("/dir0/")
        assert cliiii.isfile("/dir0/file0")
        assert cliiii.isdir("/")

    def test_mkdir(self):
        # root user
        assert not client.isdir("dir1")
        assert not client.isdir("/dir2/")
        client.mkdir("dir1")
        client.mkdir("/dir2/")
        assert client.isdir("/dir1/")
        assert client.isdir("dir2")
        # mkdir if exists
        self.assertRaises(FileExistsError, client.mkdir, "dir2/")
        # hyb user
        self.assertRaises(PermissionError, cliiii.mkdir, "dir3/")

    def test_mkdir_if_exist(self):
        # root user
        assert not client.isdir("dir4")
        client.mkdir("/dir4/")
        assert client.isdir("dir4")
        client.mkdir_if_not_exist("dir4")
        assert client.isdir("dir4")
        # hyb user
        cliiii.mkdir_if_not_exist("dir4/")

    def test_upload(self):
        # root user
        assert not client.isexist("dir0/test_upload")
        fd = open(__file__, "rb")
        client.upload(fd, "dir0", "test_upload")
        fd.close()
        assert client.isexist("dir0/test_upload")
        # hyb user

    def test_download(self):
        # root user
        fd = open("./tmp/test_download", "wb")
        client.download(fd, "dir0", "file0")
        fd.close()
        assert os.path.isfile("./tmp/test_download")
        # hyb user



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-12-07
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

from angel_worker.core.tail import Tail
from threading import Thread, current_thread
from traceback import format_exc
import requests
import json
import os


class LogCollector(Thread):
    """
    Collecting real-time log and send to the LogMonitor
    """

    def __init__(self, logger, url, dirpath, loop_time=1, last_bytes=1024):
        self.logger = logger
        self.url = url
        self.dirpath = dirpath
        self.loop_time = loop_time
        self.last_bytes = last_bytes
        self.tails = {}

    def add_log(self, task_id, filename):
        """
        Add a new log file to monitor in real-time
        """
        # check if already exist
        if task_id in self.tails:
            return
        filepath = os.path.join(self.dirpath, filename)
        # check if log file exist
        if not os.path.isfile(filepath):
            return
        tail = Tail(filepath, self.loop_time, self.last_bytes)
        tail.register_callback(self._send_log)
        self.tails[task_id] = tail
        thread = Thread(name="tasklog(%s)" % task_id)
        thread.run = tail.serve
        thread.task_id = task_id
        thread.tail = tail
        thread.start()

    def stop_logger(self, task_id=None):
        """
        Stop all log tailer
        """
        if task_id is not None:
            tail = self.tails.get(task_id)
            if tail is not None:
                tail.stop_serve()
                del self.tails[task_id]
            return
        for tail in self.tails.values():
            tail.stop_serve()
        self.tails.clear()

    def _send_log(self, log):
        """
        Helper method to send log and handle the response
        """
        task_id = current_thread().task_id
        tail = current_thread().tail
        data = {"task_id": task_id, "log": log}
        try:
            resp = requests.post(self.url, data=data)
            content = str(resp.content, encoding="utf-8")
            res = json.loads(content)
            if res.get("statusCode") != 200:
                tail.stop_serve()
                del self.tails[task_id]
        except:
            tail.stop_serve()
            del self.tails[task_id]
            self.logger.warn("send log to logmonitor fail caused by " + format_exc())




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-18
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

from angel_worker.core import config_util
from angel_worker.executor import Executor
from angel_worker.authority import Authority
from angel_worker.heartbeat import Heartbeat
from angel_worker.pull import Puller
from angel_worker.api.rpc_service import RPCService
from angel_worker.core.ftp_client import FTPClient
from angel_worker.log_collector import LogCollector
from logging import config
from threading import Thread
from traceback import format_exc
import os
import logging
import time
import signal

ETC_PATH = '/etc/angel/worker/'

executor = None
puller = None
logcollector = None
heartbeat = None

def _start():
    """
    Start up the worker service
    """
    heartbeat.start()
    puller.start()

def _on_terminated(*args):
    """
    Stop the worker service
    """
    try:
        heartbeat.stop()
        executor.clean_tasks()
        logcollector.stop_logger()
        authority.logout()
    except:
        self.logger.error(format_exc())
    os.kill(os.getpid(), signal.SIGKILL)


def _on_task_finish(task_id, state):
    """
    Callback when a task finished(success or fail)
    """
    logcollector.stop_logger(task_id)
    heartbeat.running_tasks = executor.running_tasks

def serve_forever(debug=False):
    global executor
    global puller
    global logcollector
    global heartbeat
    # setup logger
    logger_config_path = os.path.join(ETC_PATH, "logger.conf")
    logging.config.fileConfig(logger_config_path)
    if debug:
        logger = logging.getLogger("debug")
    else:
        logger = logging.getLogger("release")
    # load setting
    config_path = os.path.join(ETC_PATH, "worker.conf")
    try:
        config = config_util.parse(config_path, "worker")
        pull_time = int(config["pull_time"])
        process_refresh_time = int(config["process_refresh_time"])
        max_tasks = int(config["max_tasks"])
        script_path = config["local_script_path"]
        log_path = config["local_log_path"]
        heartbeat_time = int(config["heartbeat_time"])
        rpc_url = config["rpc_url"]
        auth_path = config["auth_path"]
        fs_host = config["fs_host"]
        fs_port = int(config["fs_port"])
        fs_timeout = int(config["fs_timeout"])
        fs_name = config["fs_name"]
        fs_pwd = config["fs_password"]
        fs_buffer_size = int(config["fs_buffer_size"])
        fs_script_path = config["fs_script_path"]
        fs_log_path = config["fs_log_path"]
        log_url = config["log_url"]
        desc = config.get("description")
    except Exception as e:
        logger.error("config file '%s' error: %s" % (config_path, str(e)))
        raise
    # create directorys
    if not os.path.isdir(script_path):
        os.makedirs(script_path)
    if not os.path.isdir(log_path):
        os.makedirs(log_path)
    # init rpc client
    rpc_service = RPCService(logger, rpc_url)
    # init filesystem client
    fs_client = FTPClient(fs_host, fs_port, fs_timeout, fs_name, fs_pwd, fs_buffer_size)
    # init authority
    authority = Authority(rpc_service, auth_path, desc, logger)
    # init executor
    executor = Executor(rpc_service, logger, process_refresh_time, max_tasks,
                             script_path, log_path, fs_client, fs_script_path, fs_log_path)
    executor.set_task_listener(_on_task_finish)
    # init heartbeat
    heartbeat = Heartbeat(rpc_service, logger, heartbeat_time)
    # init logcollector
    logcollector = LogCollector(logger, log_url, log_path)
    # init puller
    puller = Puller(logger, rpc_service, pull_time, executor, logcollector)
    # register function to catch SIGTERM signal
    signal.signal(signal.SIGTERM, _on_terminated)
    # login
    authority.login()
    # start services
    _start()




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-12-03
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

from ftplib import FTP, error_perm
import os


def ftp_decorator(func):
    """
    Decorator for initialize ftp connection
    """
    def wrapper(self, *args):
        try:
            ftp = FTP()
            ftp.connect(self.host, self.port, self.timeout)
            ftp.login(self.name, self.password)
            result = func(self, ftp, *args)
            return result
        except error_perm as e:
            raise convert_exc(e, args)
        finally:
            try:
                ftp.quit()
            except:
                pass
    return wrapper

def convert_exc(exc, args):
    """
    Help method for convert ftp response error to pylib error
    """
    msg = str(exc).lower()
    if "550" not in msg:
        return exc
    if "privileges" in msg:
        return PermissionError("no permission to access file/dir in ftp: %s" % str(args))
    elif "exists" in msg:
        return FileExistsError("file/dir already exist in ftp: %s" % str(args))
    elif "no such" in msg or "not a directory" in msg:
        return FileNotFoundError("file/dir not exists in ftp: %s" % str(args))
    else:
        return Exception(msg + str(args))


class FTPClient(object):
    """
    FTP client for crud file/dir in ftp filesystem
    """

    def __init__(self, host, port, timeout, name, password, buffer_size):
        self.host = host
        self.port = port
        self.timeout = timeout
        self.name = name
        self.password = password
        self.buffer_size = buffer_size

    def _get_info(self, ftp, path):
        """
        Obtain the information of the file/dir by the given [path]
        """
        if path.endswith("/"):
            path = path[0:-1]
        dirname = os.path.dirname(path)
        basename = os.path.basename(path)
        result = []
        try:
            ftp.dir(dirname, result.append)
            for item in result:
                if item.split(" ")[-1] == basename:
                    return item
        except error_perm as e:
            exc = convert_exc(e)
            if isinstance(exc, FileNotFoundError):
                return None
            raise exc

    @ftp_decorator
    def isfile(self, ftp, path):
        """
        Check if it's a file
        """
        info = self._get_info(ftp, path)
        if info:
            return info.startswith("-")
        return False

    @ftp_decorator
    def isdir(self, ftp, path):
        """
        Check if it's a directory
        """
        if path == "" or path =="/":
            return True
        info = self._get_info(ftp, path)
        if info:
            return info.startswith("d")
        return False

    def _isexist(self, ftp, path):
        """
        Help method for checking if the file/dir is exist
        """
        try:
            ftp.dir(path)
            return True
        except Exception as e:
            exc = convert_exc(e)
            if isinstance(exc, FileNotFoundError):
                return False
            raise exc

    @ftp_decorator
    def isexist(self, ftp, path):
        """
        Check if it's exist
        """
        return self._isexist(ftp, path)

    @ftp_decorator
    def mkdir(self, ftp, path):
        """
        Make a directory
        raise a exception if already exist
        """
        ftp.mkd(path)

    @ftp_decorator
    def mkdir_if_not_exist(self, ftp, path):
        """
        Make a directory by the given 'path' in ftp
        """
        if not self._isexist(ftp, path):
            ftp.mkd(path)

    @ftp_decorator
    def download(self, ftp, localfile, dirname, filename):
        """
        Download a file from ftp
        """
        command = "RETR " + os.path.join(dirname, filename)
        ftp.retrbinary(command, localfile.write, self.buffer_size)

    @ftp_decorator
    def read(self, ftp, func, dirname, filename):
        """
        Read file in ftp server, callback 'fun' when buffer full
        """
        command = "RETR " + os.path.join(dirname, filename)
        ftp.retrbinary(command, func, self.buffer_size)

    @ftp_decorator
    def upload(self, ftp, localfile, dirname, filename):
        """
        Upload a file to ftp
        """
        command = "STOR " + os.path.join(dirname, filename)
        ftp.storbinary(command, localfile, self.buffer_size)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2018-01-13
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93


if __name__ == "__main__":
    from angel_worker.authority import Authority
    from angel_worker.core import config_util
    from angel_worker.api.rpc_service import RPCService
    import os
    config_path = os.path.join("/etc/angel/worker", "worker.conf")
    config = config_util.parse(config_path, "worker")
    auth_path = config["auth_path"]
    desc = config["description"]
    rpc_url = config["rpc_url"]
    rpc_client = RPCService(None, rpc_url)
    auth = Authority(rpc_client, auth_path, desc)
    name = "root"
    password = "root"
    auth.register(name, password)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-12-06
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

import time
import os
import sys
from threading import Thread


class Tail(object):
    """
    Represents 'tail' command in unix
    Used to monitor log file
    """

    def __init__(self, filepath, loop_time, last_bytes):
        super(Tail, self).__init__()
        self.filepath = filepath
        self.loop_time = loop_time
        self.last_bytes = last_bytes
        self.callback = sys.stdout.write
        self.started = False

    def serve(self):
        """
        Monitor the file
        Callback begin with the last part of the file(last bytes),
        continue with every new lines
        """
        self.started = True
        filesize = os.path.getsize(self.filepath)
        with open(self.filepath, "r") as fd:
            # seek to the last part of the file
            if filesize > self.last_bytes:
                fd.seek(filesize - self.last_bytes, os.SEEK_SET)
            # read the end lines of the file
            while self.started:
                lines = fd.readlines()
                if lines:
                    self._handle_lines(lines)
                else:
                    time.sleep(self.loop_time)

    def _handle_lines(self, lines):
        """
        Callback when new lines received
        """
        # conbine mulit lines into a message
        if not lines:
            return
        if len(lines) == 1:
            result = lines[0]
        elif len(lines) > 1:
            result = ""
            for line in lines:
                result += line
        # callback with new lines
        if self.callback:
            self.callback(result)

    def register_callback(self, callback):
        """
        Set callback function
        """
        self.callback = callback

    def stop_serve(self):
        """
        Stop the Tail
        """
        self.started = False



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-21
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

import hashlib
import os
import sys


class Authority(object):
    """
    Helper class for register, login and logout
    """
    def __init__(self, rpc_client, auth_path, desc, logger=None):
        self.rpc_client = rpc_client
        self.auth_path = auth_path
        self.desc = desc
        self.logger = logger

    def register(self, name, password, force=False):
        """
        Send a register request to master with 'name' and 'password'
        External command for shell, not deamon
        """
        if not force and os.path.isfile(self.auth_path):
            raise Exception("already register")
        # obtain the hash of name
        _name = hashlib.sha1(name.encode("utf8")).hexdigest()
        # obtain the hash of password
        _password = hashlib.sha1(password.encode("utf8")).hexdigest()
        # register by rpc client
        params = {}
        if self.desc:
            params["desc"] = self.desc
        resp = self.rpc_client.register(_name, _password, params)
        # check if login successfully
        if resp.success():
            worker_id = resp.content
            # store key(worker_id, name, password) after register
            with open(self.auth_path, "w") as f:
                f.write(str(worker_id))
                f.write("\n")
                f.write(_name)
                f.write("\n")
                f.write(_password)
            sys.stdout.write("register successfully, author file : %s\n" % self.auth_path)
        else:
            raise Exception(resp.content)

    def login(self):
        """
        Send a login request to master
        Return worker_id if login successfully
        """
        # check if the auth file exist
        if not os.path.isfile(self.auth_path):
            # no auth file, not register before
            message = "login fail, not register before login"
            self.logger.warn(message)
            raise Exception(message)
        with open(self.auth_path, "r") as f:
            # read auth file, get the hash of name and password
            key = f.read()
            keys = key.split("\n")
            if len(keys) != 3:
                self.logger.error("something wrong with the key in %s" % self.auth_path)
            try:
                # login by rpc client
                resp = self.rpc_client.login(keys[1], keys[2], int(keys[0]))
                if resp.success():
                    self.logger.info("login successfully")
                    self.rpc_client.worker_id = int(keys[0])
                else:
                    message = "login fail, Master: %s" % resp.content
                    self.logger.warn(message)
                    raise Exception (message)
            except Exception as e:
                message = "login fail caused by %s" % str(e)
                self.logger.warn(message)
                raise Exception(message)

    def logout(self):
        """
        Send a logout request to master
        """
        try:
            resp = self.rpc_client.logout()
            if resp.success():
                self.logger.info("logout successfully")
                return
            else:
                message = "logout fail, Master: %s" % resp.content
        except Exception as e:
            message = "logout fail, caused by %s" % str(e)
        raise Exception(message)




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-21
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

from angel_worker.api.response import ResponseMessage
from xmlrpc.client import ServerProxy
from traceback import format_exc
import time
from http.client import CannotSendRequest


def exception_handler(func):
    """
    Handle exception and package the response
    """
    def wrapper(self, *args, **kwargs):
        try:
            resp = func(self, *args, **kwargs)
            return ResponseMessage(resp.get("code"), resp.get("content"))
        except AttributeError as e:
            self.logger.warn("%s, maybe not not login before" % str(e))
            raise
        except Exception as e:
            self.logger.warn("%s rpc fail caused by %s" % (func.__name__, format_exc()))
            raise
    return wrapper


class RPCService(object):
    """
    RPC Client that call the service of master
    """

    def __init__(self, logger, url):
        self.url = url
        self.logger = logger

    @exception_handler
    def register(self, name, password, params):
        client = ServerProxy(self.url)
        return client.register(name, password, params)

    @exception_handler
    def login(self, name, password, worker_id):
        client = ServerProxy(self.url)
        return client.login(name, password, worker_id)

    @exception_handler
    def logout(self):
        client = ServerProxy(self.url)
        return client.logout(self.worker_id)

    @exception_handler
    def heartbeat(self, params):
        if not hasattr(self, "heartbeat_client"):
            self.heartbeat_client = ServerProxy(self.url)
        return self.heartbeat_client.heartbeat(self.worker_id, params)

    @exception_handler
    def pull_tasks(self):
        if not hasattr(self, "pulltask_client"):
            self.pulltask_client = ServerProxy(self.url)
        return self.pulltask_client.pull_tasks(self.worker_id)

    @exception_handler
    def pull(self):
        if not hasattr(self, "pull_client"):
            self.pull_client = ServerProxy(self.url)
        return self.pull_client.pull(self.worker_id)

    @exception_handler
    def task_callback(self, task_id, state):
        client = ServerProxy(self.url)
        return client.task_callback(task_id, int(time.time()), self.worker_id, state)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2018-01-12
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

import time
import os
import signal
from threading import Thread
from traceback import format_exc

CAMMAND = "cmd"
TASKS = "tasks"
LOGS = "logs"

class Puller(Thread):
    """
    A looper thread for pulling commands from master
    """

    def __init__(self, logger, rpc_service, pull_time, executor, logcollector):
        super(Puller, self).__init__()
        self.logger = logger
        self.rpc_service = rpc_service
        self.pull_time = pull_time
        self.executor = executor
        self.logcollector = logcollector
        self.started = False

    def run(self):
        """
        Main loop of the pulling thread
        """
        self.started = True
        while self.started:
            # sleep for a while
            time.sleep(self.pull_time)
            # obtain command from master
            try:
                resp = self.rpc_service.pull()
                if resp.success():
                    if resp.content:
                        self.logger.debug(resp.content)
                    cmd = resp.content.get(CAMMAND)
                    tasks = resp.content.get(TASKS)
                    logs = resp.content.get(LOGS)
                else:
                    self.logger.warn("pull fail caused by %s" % resp.content)
                    continue
            except Exception as e:
                self.logger.warn(format_exc())
                continue
            if cmd == 'shutdown':
                os.kill(os.getpid(), signal.SIGTERM)
                return
            try:
                if tasks:
                    self.executor.submit(tasks)
                if logs:
                    for log in logs:
                        self.logcollector.add_log(logid, str(logid))
            except:
                self.logger.error(format_exc())

        def stop(self):
            """
            Stop the puller
            """
            self.started = False




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-18
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

from threading import Thread
import psutil
import time
import os
import signal


class Heartbeat(Thread):
    """
    A heartbeat thread that get its performance and sent to the master
    """
    def __init__(self, rpc_client, logger, heartbeat_time):
        super(Heartbeat, self).__init__()
        self.started = False
        self.heartbeat_time = heartbeat_time
        self.rpc_client = rpc_client
        self.logger = logger
        self.running_tasks = 0

    def start(self):
        """
        start up the heartbeat thread
        """
        self.started = True
        super(Heartbeat, self).start()

    def stop(self):
        """
        stop the heartbeat thread
        """
        self.started = False

    def run(self):
        """
        Main loop of the heartbeat thread
        """
        performances = {}
        while self.started:
            # sleep for a while
            time.sleep(self.heartbeat_time)
            # get its performance parameters
            cpu_free = 100 - psutil.cpu_percent()
            mem_free = psutil.virtual_memory().available
            disk_io = psutil.disk_io_counters()
            net_io = psutil.net_io_counters()
            performances["cpu_free"] = cpu_free
            performances["mem_free"] = mem_free >> 20
            performances["disk_write"] = disk_io.write_count
            performances["disk_read"] = disk_io.read_count
            performances["net_send"] = net_io.packets_sent
            performances["net_rev"] = net_io.packets_recv
            performances["running_tasks"] = self.running_tasks
            # send to the master
            try:
                resp = self.rpc_client.heartbeat(performances)
                if not resp.success():
                    self.logger.warn("heartbeat error caused by:" % resp.content)
                    os.kill(os.getpid(), signal.SIGTERM)
            except:
                pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-21
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

CODE_SUCCESS = 0
CODE_FAIL = 1


class ResponseMessage(object):
    """
    Response message from master when rpc
    """

    def __init__(self, code=CODE_SUCCESS, content=None):
        self.code = code
        self.content = content

    def success(self):
        return self.code == CODE_SUCCESS



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-18
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

STATE_INIT = 100
STATE_TIMING = 101
STATE_RUNNING = 102
STATE_SCHEDULE_TIMEOUT = 103
STATE_RUNNING_TIMEOUT = 104
STATE_DOWNLOAD_FAIL = 105
STATE_SUCCESS = 0
STATE_FAIL = 1

ALL_STATES = set([STATE_INIT, STATE_TIMING, STATE_RUNNING, STATE_SCHEDULE_TIMEOUT,
                  STATE_RUNNING_TIMEOUT, STATE_DOWNLOAD_FAIL, STATE_SUCCESS, STATE_FAIL])

#INIT_STATE = 0
#TIMING_STATE = 1
#SUCCESS_STATE = 2
#RUNNING_STATE = 3
#SCHEDULE_TIMEOUT_STATE = 4
#RUNNING_TIMEOUT_STATE = 5
#FAIL = 6



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-18
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93

from collections import namedtuple
from concurrent.futures.thread import ThreadPoolExecutor
from angel_worker import constants
from traceback import format_exc
import os, stat, subprocess, time

Task = namedtuple("Task", "id job_id state")



class Executor(object):
    """
    A executor that execute the tasks obtained from master
    """

    def __init__(self, rpc_client, logger, process_refresh_time, max_tasks,
                 script_path, log_path, fs_client, fs_script_path, fs_log_path):
        self.started = False
        self.rpc_client = rpc_client
        self.logger = logger
        self.process_refresh_time = process_refresh_time
        self.thread_pool = ThreadPoolExecutor(max_tasks)
        self.script_path = script_path
        self.log_path = log_path
        self.fs_client = fs_client
        self.fs_script_path = fs_script_path
        self.fs_log_path = fs_log_path
        self.processes = {}
        self.running_tasks = 0
        self.task_listener = None
        # create logs and scripts directory
        if not os.path.isdir(log_path):
            os.mkdir(log_path)
        if not os.path.isdir(script_path):
            os.mkdir(script_path)

    def clean_tasks(self):
        """
        Stop all tasks
        """
        for process in self.processes.values():
            process.kill()

    def set_task_listener(self, listener):
        """
        setup a listener for changing state of tasks
        """
        self.task_listener = listener

    def submit(self, tasks):
        """
        Submit tasks to executor
        """
        for item in tasks:
            task = Task(**item)
            if task.state == constants.STATE_RUNNING_TIMEOUT:
                # stop the running timeout tasks
                process = self.processes.get(task.id)
                if process:
                    process.kill()
            else:
                # new processes to execute tasks
                self.thread_pool.submit(self._run_task, task)

    def _run_task(self, task):
        """
        run the task in the new process
        """
        # download the script if not exist in local file system
        path = os.path.join(self.script_path, "%d" % task.job_id)
        try:
            # download from file system
            with open(path, "wb") as fd:
                self.fs_client.download(fd, self.fs_script_path, str(task.job_id))
        except Exception as e:
            os.remove(path)
            self.logger.warn("download scipt file fail caused by %s" % str(e))
            try:
                self.rpc_client.task_callback(task.id, constants.STATE_DOWNLOAD_FAIL)
            except Exception as e:
                self.logger.warn("task callback fail caused by %s" % str(e))
            return
        # add exec flag
        os.chmod(path, stat.S_IRWXU)
        # open files to store the logs of the process
        with open(os.path.join(self.log_path, str(task.id)), "w") as f:
            self.running_tasks = self.running_tasks + 1
            # fork a new process to execute the script
            p = subprocess.Popen(path, stdout=f, stderr=f)
            self.processes[task.id] = p
            # check if it is finished
            while p.poll() is None:
                time.sleep(self.process_refresh_time)
        # upload the log
        try:
            with open(os.path.join(self.log_path, "%d" % task.id), "rb") as f:
                self.fs_client.upload(f, self.fs_log_path, str(task.id))
        except Exception as e:
            self.logger.error(format_exc())
        # remove the process
        del self.processes[task.id]
        # callback the master that the task is finished(success or fail)
        returncode = p.poll()
        self.running_tasks = self.running_tasks - 1
        if returncode == 0:
            state = constants.STATE_SUCCESS
        elif returncode == -9:
            state = constants.STATE_RUNNING_TIMEOUT
        else:
            state = constants.STATE_FAIL
        if self.task_listener:
            self.task_listener(task.id, state)
        try:
            self.rpc_client.task_callback(task.id, state)
        except Exception as e:
            self.logger.warn("task callback fail caused by %s" % format_exc())




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2018-01-10
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93


from setuptools import setup, find_packages
import sys
import os

ETC_PATH = "/etc/angel/worker/"

setup(
    name="angel_worker",
    version="1.0.5",
    author="Lafite93",
    author_email="2273337844@qq.com",
    description="The worker of a distribute job scheduler named 'angel'",
    license="Apache",
    keywords="angel worker executor",
    url="https://pypi.python.org/pypi/angel_worker",
    packages=find_packages(),
    include_package_data=True,
    data_files=[
        (ETC_PATH, ['etc/worker.conf', 'etc/logger.conf'])
    ],
    scripts=["scripts/worker_daemon", "scripts/register_worker"],
    install_requires=[
        "configparser",
        "psutil",
        "requests",
        "daemonpy3"
    ],
    zip_safe=False,
    platforms=["Linux", "Unix"],
    classifiers = [
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.5',
    ],
    entry_points={
        "console_scripts":[
            "start_worker = angel_worker.worker:serve_forever"
        ]
    }
)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# @Time         2017-11-30
# @Email        2273337844@qq.com
# @Copyright    © 2017 Lafite93


if __name__ == "__main__":
    from angel_worker import worker
    worker.serve_forever(debug=True)
